import statistics
from .ortega import ORTEGA
import math
from .common import __timedifcheck
from .common import __check_dist
import pandas as pd
# from osgeo import ogr


def compute_ppa_size(interation: ORTEGA):
    size_list1 = [e.el[0].length for e in interation.ellipses_list_id1]
    size_list2 = [e.el[0].length for e in interation.ellipses_list_id2]
    id1 = interation.id1
    id2 = interation.id2

    ellipse_size_collection = {"size_list1": size_list1, "size_list2": size_list2}
    print(f"Statistics of PPA ellipses length for id {id1}:")
    print(f"mean:", statistics.mean(ellipse_size_collection['size_list1']))
    print(f"min:", min(ellipse_size_collection['size_list1']))
    print(f"max:", max(ellipse_size_collection['size_list1']))
    print(f"median:", statistics.median(ellipse_size_collection['size_list1']))
    print(f"std:", statistics.stdev(ellipse_size_collection['size_list1']))
    print(f"Statistics of PPA ellipses length for id {id2}:")
    print(f"mean:", statistics.mean(ellipse_size_collection['size_list2']))
    print(f"min:", min(ellipse_size_collection['size_list2']))
    print(f"max:", max(ellipse_size_collection['size_list2']))
    print(f"median:", statistics.median(ellipse_size_collection['size_list2']))
    print(f"std:", statistics.stdev(ellipse_size_collection['size_list2']))
    return ellipse_size_collection


def compute_ppa_interval(interation: ORTEGA):
    id1 = interation.id1
    id2 = interation.id2
    print(f"Statistics of PPA ellipses time interval (minutes) for id {id1}:")
    time_diff = [
        interation.df1[interation.time_field].diff().dt.total_seconds().div(60).dropna(),
        interation.df2[interation.time_field].diff().dt.total_seconds().div(60).dropna()
    ]
    print(f"mean:", time_diff[0].mean())
    print(f"min:", time_diff[0].min())
    print(f"max:", time_diff[0].max())
    print(f"median:", time_diff[0].median())
    print(f"std:", time_diff[0].std())

    print(f"Statistics of PPA ellipses time interval (minutes) for id {id2}:")
    print(f"mean:", time_diff[1].mean())
    print(f"min:", time_diff[1].min())
    print(f"max:", time_diff[1].max())
    print(f"median:", time_diff[1].median())
    print(f"std:", time_diff[1].std())
    return time_diff


def compute_ppa_speed(df: pd.DataFrame):
    def compute_speed(x):
        d1 = __check_dist(x['P1_startlat'], x['P1_startlon'], x['P1_endlat'], x['P1_endlon'])
        delta_time1 = __timedifcheck(x['P1_t_start'], x['P1_t_end'])
        x['P1_speed'] = d1 / delta_time1  # speed in m/s
        d2 = __check_dist(x['P2_startlat'], x['P2_startlon'], x['P2_endlat'], x['P2_endlon'])
        delta_time2 = __timedifcheck(x['P2_t_start'], x['P2_t_end'])
        x['P2_speed'] = d2 / delta_time2  # speed in m/s
        return x

    df = df.apply(lambda x: compute_speed(x), axis=1)
    df['diff_speed'] = (df['P2_speed'] - df['P1_speed']).abs() / (
                (df['P1_speed'] + df['P2_speed']) / 2)
    print("Statistics of percentage difference in movement speed between intersecting PPAs:")
    print(df['diff_speed'].describe())
    return df


def compute_ppa_direction(df: pd.DataFrame):
    def compute_direction(x):
        # Angle between p1 and p2 in degree, Difference in x coordinates, Difference in y coordinates
        x['P1_direction'] = math.degrees(math.atan2(x['P1_endlon'] - x['P1_startlon'], x['P1_endlat'] - x['P1_startlat']))
        x['P2_direction'] = math.degrees(math.atan2(x['P2_endlon'] - x['P2_startlon'], x['P2_endlat'] - x['P2_startlat']))
        return x

    def between_angles(x, a):
        if x[a] >= 180:
            x[a] -= 360
        if x[a] < -180:
            x[a] += 360
        x[a] = abs(x[a])
        return x

    df = df.apply(lambda x: compute_direction(x), axis=1)
    df['diff_direction'] = df['P2_direction'] - df['P1_direction']
    df = df.apply(lambda x: between_angles(x, 'diff_direction'), axis=1)
    print("Statistics of difference in movement direction between intersecting PPAs:")
    print(df['diff_direction'].describe())
    return df

# def output_shapefile(interation: ORTEGA):
#     # Now convert it to a shapefile with OGR
#     driver = ogr.GetDriverByName('Esri Shapefile')
#     id1 = interation.id1
#     id2 = interation.id2
#     ds = driver.CreateDataSource(f"{id1}_{id2}.shp")
#     layer = ds.CreateLayer('', None, ogr.wkbPolygon)
#     # Add one attribute
#     layer.CreateField(ogr.FieldDefn('pid', ogr.OFTInteger))
#     layer.CreateField(ogr.FieldDefn('lat', ogr.OFTReal))
#     layer.CreateField(ogr.FieldDefn('lon', ogr.OFTReal))
#     layer.CreateField(ogr.FieldDefn('time', ogr.OFTString))
#     layer.CreateField(ogr.FieldDefn('last_lat', ogr.OFTReal))
#     layer.CreateField(ogr.FieldDefn('last_lon', ogr.OFTReal))
#     layer.CreateField(ogr.FieldDefn('last_time', ogr.OFTString))
#     layer.CreateField(ogr.FieldDefn('ppa_id', ogr.OFTInteger))
#     defn = layer.GetLayerDefn()
#
#     i = 0
#     for item in interation.ellipses_list:
#         if abs(pd.Timedelta(item.t2 - item.t1).total_seconds()) > interation.max_el_time_min * 60:
#             continue
#         # Create a new feature (attribute and geometry)
#         feat = ogr.Feature(defn)
#         i += 1
#         feat.SetField('pid', item.pid)
#         feat.SetField('lat', item.lat)
#         feat.SetField('lon', item.lon)
#         feat.SetField('last_lat', item.last_lat)
#         feat.SetField('last_lon', item.last_lon)
#         feat.SetField('time', str(item.t1))
#         feat.SetField('last_time', str(item.t2))
#         # feat.SetField('time', item.t1.year, item.t1.month, item.t1.day, item.t1.hour, item.t1.minute, item.t1.second, 0)
#         # feat.SetField('last_time', item.t2.year, item.t2.month, item.t2.day, item.t2.hour, item.t2.minute, item.t2.second, 0)
#         feat.SetField('ppa_id', i)
#
#         # Make a geometry, from Shapely object
#         geom = ogr.CreateGeometryFromWkb(item.geom.wkb)
#         feat.SetGeometry(geom)
#         layer.CreateFeature(feat)
#         feat = geom = None  # destroy these
#     # Save and close everything
#     ds = layer = feat = geom = None


import geopandas as gpd


def points_to_traj(traj_points, col=['Lng', 'Lat', 'ID'], timecol=None):
    '''
    Input trajectory, generate GeoDataFrame
    Parameters
    -------
    traj_points : DataFrame
        trajectory data
    col : List
        The column name, in the sequence of [lng, lat,trajectoryid]
    timecol : str(Optional)
        Optional, the column name of the time column. If given, the geojson
        with [longitude, latitude, altitude, time] in returns can be put into
        the Kepler to visualize the trajectory
    Returns
    -------
    traj : GeoDataFrame
        Generated trajectory
    '''
    [Lng, Lat, ID] = col
    if timecol:
        geometry = []
        traj_id = []
        for i in traj_points[ID].drop_duplicates():
            coords = traj_points[traj_points[ID] == i][[Lng, Lat, timecol]]
            coords[timecol] = coords[timecol].apply(
                lambda r: int(r.value / 1000000000))
            coords['altitude'] = 0
            coords = coords[[Lng, Lat, 'altitude', timecol]].values.tolist()
            traj_id.append(i)
            if len(coords) >= 2:
                geometry.append({
                    "type": "Feature",
                    "properties": {"ID": i},
                    "geometry": {"type": "LineString",
                                 "coordinates": coords}})
        traj = {"type": "FeatureCollection",
                "features": geometry}
    else:
        traj = gpd.GeoDataFrame()
        from shapely.geometry import LineString
        geometry = []
        traj_id = []
        for i in traj_points[ID].drop_duplicates():
            coords = traj_points[traj_points[ID] == i][[Lng, Lat]].values
            traj_id.append(i)
            if len(coords) >= 2:
                geometry.append(LineString(coords))
            else:
                geometry.append(None)  # pragma: no cover
        traj[ID] = traj_id
        traj['geometry'] = geometry
        traj = gpd.GeoDataFrame(traj)
    return traj



import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
from .ortega import ORTEGA
from .ellipses import Ellipse
from typing import List, Tuple
import matplotlib.patches as mpatches

# from .traj import *
# from keplergl import KeplerGl
# from matplotlib.animation import FuncAnimation


def plot_original_tracks(interation: ORTEGA, throw_out_big_ellipses: bool = True, legend: bool = True,
                         save_plot: bool = False):
    color1 = "#ff0000"
    color2 = "#0000ff"
    colors = [color1, color2]
    fig = plt.figure(1, figsize=(8, 8), dpi=90)
    # fig.set_tight_layout(True)

    for i, collection in enumerate([interation.ellipses_list_id1, interation.ellipses_list_id2]):
        color_picked = colors[i]
        for item in collection:
            if throw_out_big_ellipses and abs(
                    pd.Timedelta(item.t2 - item.t1).total_seconds()) >= interation.max_el_time_min * 60:
                continue
            # if throw_out_big_ellipses and item.el[0].length > max_val:
            #     continue
            x, y = item.el[0].xy
            plt.plot(y, x, color=color_picked, alpha=0.8, linewidth=1, solid_capstyle="round")
            # PLOT THE POINTS USED TO MAKE THE ELLIPSES, TOO
            plt.plot([item.lon, item.last_lon], [item.lat, item.last_lat], "o-", color="grey", linewidth=0.5,
                     markersize=1)

    # plt.title('')
    plt.xlabel("x", fontsize=14)
    plt.ylabel("y", fontsize=14)
    plt.grid(True)
    if legend:
        color1_patch = mpatches.Patch(color=color1, label=interation.id1)
        color2_patch = mpatches.Patch(color=color2, label=interation.id2)
        plt.legend(handles=[color1_patch, color2_patch])
    if save_plot:
        plt.savefig(f"{interation.id1}_{interation.id2}_original_tracks.pdf")
    plt.show()


def plot_interaction(interation: ORTEGA, all_intersection_pairs: List[Tuple[Ellipse, Ellipse]], throw_out_big_ellipses: bool = True,
                     legend: bool = True, save_plot: bool = False):
    color1 = "#ff0000"  # "#3ABA36" #green #
    color2 = "#0000ff"  # "#ff0000" #red %"#005EFF" #blue #"
    interaction_color = "#f2ff00"  # yellow
    colors = [color1, color2]
    fig = plt.figure(1, figsize=(8, 8), dpi=90)
    # fig.set_tight_layout(True)

    for i, collection in enumerate([interation.ellipses_list_id1, interation.ellipses_list_id2]):
        color_picked = colors[i]
        for item in collection:
            if throw_out_big_ellipses and abs(
                    pd.Timedelta(item.t2 - item.t1).total_seconds()) >= interation.max_el_time_min * 60:
                continue
            # if throw_out_big_ellipses and item.el[0].length > max_val:
            #     continue
            x, y = item.el[0].xy
            plt.plot(y, x, color=color_picked, alpha=0.8, linewidth=1, solid_capstyle="round")
            # PLOT THE POINTS USED TO MAKE THE ELLIPSES, TOO
            plt.plot([item.lon, item.last_lon], [item.lat, item.last_lat], "o-", color="grey", linewidth=0.5,
                     markersize=1)

    for two_item in all_intersection_pairs:
        for item in two_item:
            if throw_out_big_ellipses and abs(
                    pd.Timedelta(item.t2 - item.t1).total_seconds()) >= interation.max_el_time_min * 60:
                continue
            x1, y1 = item.el[0].xy
            plt.plot(y1, x1, color=interaction_color, alpha=0.8, linewidth=1, solid_capstyle="round")
            plt.plot([item.lon, item.last_lon], [item.lat, item.last_lat], "o-", color="grey", linewidth=0.5,
                     markersize=1)
    # plt.title('')
    plt.xlabel("x", fontsize=14)
    plt.ylabel("y", fontsize=14)
    plt.grid(True)
    if legend:
        color1_patch = mpatches.Patch(color=color1, label=interation.id1)
        color2_patch = mpatches.Patch(color=color2, label=interation.id2)
        plt.legend(handles=[color1_patch, color2_patch])
    if save_plot:
        plt.savefig(f"{interation.id1}_{interation.id2}_interaction.pdf")
    plt.show()


def visualization_trip(interaction, zoom='auto', height=500):
    '''
    The input is the trajectory data and the column name. The output is the
    visualization result based on kepler
    Parameters
    -------
    interaction :

    zoom : number
        Map zoom level
    height : number
        The height of the map frame
    Returns
    -------
    vmap : keplergl.keplergl.KeplerGl
        Visualizations provided by keplergl
    '''
    print('Processing movement data...')
    Lng, Lat, ID, timecol = interaction.longitude_field, interaction.latitude_field, interaction.id_field, interaction.time_field

    if interaction.start_time is not None or interaction.end_time is not None:
        trajdata = interaction.subset
    else:
        trajdata = interaction.data

    # clean data
    trajdata = trajdata[-((trajdata[Lng].isnull()) | (trajdata[Lat].isnull()))]
    trajdata = trajdata[
        (trajdata[Lng] >= -180) & (trajdata[Lng] <= 180) & (trajdata[Lat] >= -90) & (trajdata[Lat] <= 90)]

    trajdata[timecol] = pd.to_datetime(trajdata[timecol])

    trajdata = trajdata.sort_values(by=[ID, timecol])
    traj = points_to_traj(trajdata, col=[Lng, Lat, ID], timecol=timecol)
    ls = []
    for i in range(len(traj['features'])):
        ls.append(traj['features'][i]['geometry']['coordinates'][0])
        ls.append(traj['features'][i]['geometry']['coordinates'][-1])
    ls = pd.DataFrame(ls)
    lon_center, lat_center, starttime = ls[0].mean(), ls[1].mean(), ls[3].min()
    if zoom == 'auto':
        lon_min, lon_max = ls[0].quantile(0.05), ls[0].quantile(0.95)
        zoom = 8.5 - np.log(lon_max - lon_min) / np.log(2)
    print('Generate visualization...')
    vmap = KeplerGl(config={
        "version": "v1",
        "config":
            {"visState":
                {
                    "filters": [],
                    "layers": [
                        {
                            "id": "hizm36i",
                            "type": "trip",
                            "config":
                                {
                                    "dataId": "trajectory",
                                    "label": "trajectory",
                                    # "color": [255, 255, 255],
                                    # "highlightColor": [255, 255, 0, 255],
                                    "columns":
                                        {
                                            "geojson": "_geojson"
                                        },
                                    "isVisible": True,
                                    "visConfig": {
                                        "opacity": 0.8,
                                        "thickness": 5,
                                        "colorRange": {
                                            "name": "ColorBrewer Set1-6",
                                            "type": "qualitative",
                                            "category": "ColorBrewer",
                                            "colors": [
                                                "#e41a1c",
                                                "#377eb8",
                                                "#4daf4a",
                                                "#984ea3",
                                                "#ff7f00",
                                                "#ffff33",
                                            ],
                                        },
                                        "trailLength": 1000,
                                        "sizeRange": [0, 10],
                                    },
                                },
                            "visualChannels": {
                                "colorField": {"name": "ID", "type": "integer"},
                                "colorScale": "quantile",
                                "sizeField": None,
                                "sizeScale": "linear",
                            },
                        }],
                    "layerBlending": "additive",
                    "animationConfig":
                        {
                            "currentTime": starttime,
                            "speed": 0.1
                        }
                },
                "mapState":
                    {
                        "bearing": 0,
                        "latitude": lat_center,
                        "longitude": lon_center,
                        "pitch": 0,
                        "zoom": zoom,
                    },
                "mapStyle": {
                    "styleType": "satellite"  # outdoors, streets
                }
            }},
        data={'trajectory': traj}, height=height)
    return vmap

# def plot_interaction_animated(all_intersection_pairs: List[Tuple[Ellipse, Ellipse]], ellipses_list_id1: List[Ellipse],
#                               ellipses_list_id2: List[Ellipse], boundary: [[float, float], [float, float]],
#                               id1: int, id2: int, max_el_time_min: float,
#                               throw_out_big_ellipses: bool = True, legend: bool = True, save_plot: bool = False):
#     color1 = "#ff0000"
#     color2 = "#0000ff"
#     colors = [color1, color2]
#     # fig = plt.figure(1, figsize=(8, 8), dpi=90)
#     fig, ax = plt.subplots(figsize=(8, 8))
#     fig.set_tight_layout(True)
#     plt.xlabel("Longitude", fontsize=14)
#     plt.ylabel("Latitude", fontsize=14)
#     plt.grid(True)
#     plt.xlim(boundary[0])
#     plt.ylim(boundary[1])
#     if legend:
#         color1_patch = mpatches.Patch(color=color1, label=id1)
#         color2_patch = mpatches.Patch(color=color2, label=id2)
#         plt.legend(handles=[color1_patch, color2_patch])
#
#     # print(len(ellipses_list_id1), len(ellipses_list_id2))
#     # boundary = [[min_lon - 0.01, max_lon + 0.01], [min_lat - 0.01, max_lat + 0.01]]
#
#     # G = ox.graph_from_bbox(boundary[1][1], boundary[1][0], boundary[0][1], boundary[0][0], network_type='all')
#     # ox.plot_graph(G, ax=ax, dpi=300, show=False, close=False, node_size=0, bgcolor='white', edge_color='gray',
#     #               edge_linewidth=0.5)
#
#     def animate(i, x=[], y=[]):
#         # if i % 5 == 0:
#         #     plt.cla()
#         # concerns: length of the two lists, temporal order of two lists needs to be consistent?
#         # based on days
#         min_len = min(len(ellipses_list_id1), len(ellipses_list_id2))
#         max_len = max(len(ellipses_list_id1), len(ellipses_list_id2))
#
#         if i < min_len:
#             for j, item in enumerate([ellipses_list_id1[i], ellipses_list_id2[i]]):
#                 if throw_out_big_ellipses and abs(
#                         pd.Timedelta(item.t2 - item.t1).total_seconds()) >= max_el_time_min * 60:
#                     pass
#                 else:
#                     x, y = item.el[0].xy
#                     plt.plot(y, x, color=colors[j], alpha=0.8, linewidth=1, solid_capstyle="round")
#                     plt.plot([item.lon, item.last_lon], [item.lat, item.last_lat], "o-", color="grey", linewidth=0.5,
#                              markersize=1)
#         elif i < max_len:
#             if len(ellipses_list_id1) > len(ellipses_list_id2):
#                 item = ellipses_list_id1[i]
#                 k = 0
#             else:
#                 item = ellipses_list_id2[i]
#                 k = 1
#
#             if throw_out_big_ellipses and abs(
#                     pd.Timedelta(item.t2 - item.t1).total_seconds()) >= max_el_time_min * 60:
#                 pass
#             else:
#                 x, y = item.el[0].xy
#                 plt.plot(y, x, color=colors[k], alpha=0.8, linewidth=1, solid_capstyle="round")
#                 plt.plot([item.lon, item.last_lon], [item.lat, item.last_lat], "o-", color="grey", linewidth=0.5,
#                          markersize=1)
#
#     ani = FuncAnimation(fig, animate, interval=300)
#     plt.show()
#     if save_plot:
#         ani.to_html5_video()



from setuptools import setup

if __name__ == "__main__":
    setup(
        name='ortega',
        version='0.0.31',
        author='MOVE lab@UCSB',
        author_email="rongxiangsu@ucsb.edu",
        packages=["ortega"],
        url='https://move.geog.ucsb.edu/',
        license='MIT',
        description='ORTEGA',
        install_requires=[
            "numpy",
            "pandas",
            "shapely",
            "attrs",
            "matplotlib",
            "statistics",
            "typing_extensions",
        ],
    )
